#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pyspark.sql.types import IntegerType

from petastorm.codecs import ScalarCodec, NdarrayCodec
from petastorm.ngram import NGram
from petastorm.reader_impl.row_group_decoder import RowDecoder
from petastorm.unischema import Unischema, UnischemaField, dict_to_spark_row

_matrix_shape = (10, 20)

TestSchema = Unischema('TestSchema', [
    UnischemaField('some_number', np.uint8, (), ScalarCodec(IntegerType()), False),
    UnischemaField('some_matrix', np.float32, _matrix_shape, NdarrayCodec(), False),
])


def _rand_row(some_number=np.random.randint(0, 255)):
    row_as_dict = {
        'some_number': some_number,
        'some_matrix': np.random.random(size=_matrix_shape).astype(np.float32),
    }

    return row_as_dict


def test_row_decoding():
    expected_row = _rand_row()
    encoded_row = dict_to_spark_row(TestSchema, expected_row).asDict()

    decoder = RowDecoder(TestSchema, None)
    actual_row = decoder.decode(encoded_row)._asdict()

    # Out-of-the-box pytest `assert actual_row == expected_row` can not compare dictionaries properly
    np.testing.assert_equal(actual_row, expected_row)


def test_ngram_decoding():
    N = 5
    ngram_spec = NGram({
        -1: [TestSchema.some_number, TestSchema.some_matrix],
        0: [TestSchema.some_number],
        1: [TestSchema.some_number, TestSchema.some_matrix],
    }, 2, TestSchema.some_number)

    expected_rows = [_rand_row(n) for n in range(N)]
    encoded_rows = [dict_to_spark_row(TestSchema, row).asDict() for row in expected_rows]
    encoded_ngrams = ngram_spec.form_ngram(encoded_rows, TestSchema)

    decoder = RowDecoder(TestSchema, ngram_spec)

    # decoded_ngrams is a list of 3 dictionaries, each have -1, 0, 1 keys.
    decoded_ngrams = [decoder.decode(encoded) for encoded in encoded_ngrams]

    # Verify we got 3 dictionaries
    assert 3 == len(decoded_ngrams)

    single_sample = decoded_ngrams[0]

    # A single decoded ngram looks like this:
    #   -1: some_number, some_matrix
    #    0: some_number
    #    1: some_number, some_matrix
    assert 2 == len(single_sample[-1])
    assert 0 == single_sample[-1].some_number

    assert 1 == len(single_sample[0])
    assert 1 == single_sample[0].some_number

    assert 2 == len(single_sample[1])
    assert 2 == single_sample[1].some_number



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import hashlib

import numpy as np
from pyarrow import parquet as pq
from pyarrow.parquet import ParquetFile
from six.moves.urllib.parse import urlparse, urlunparse

from petastorm.cache import NullCache
from petastorm.fs_utils import FilesystemResolver


def _merge_two_dicts(a, b):
    """Merges two dictionaries together. If the same key is present in both input dictionaries, the value from 'b'
    dominates."""
    result = a.copy()
    result.update(b)
    return result


def _select_cols(a_dict, keys):
    """Filteres out entries in a dictionary that have a key which is not part of 'keys' argument. `a_dict` is not
    modified and a new dictionary is returned."""
    if keys == list(a_dict.keys()):
        return a_dict
    else:
        return {field_name: a_dict[field_name] for field_name in keys}


class RowGroupLoader(object):
    def __init__(self, dataset_url, schema, ngram, local_cache, worker_predicate, hdfs_driver='libhdfs3'):
        """RowGroupLoader responsible for loading one rowgroup at a time. Rows returned are returned encoded.

        :param dataset_url: A url of a parquet dataset.
        :param schema: A unischema corresponding to the data in the dataset
        :param ngram: An instance of NGram if ngrams should be read or None, if each row in the dataset corresponds to
          a single sample returned.
        :param local_cache: An instance of a rowgroup cache (CacheBase interface) object to be used.
        :param worker_predicate: An instance of predicate (PredicateBase interface)
        :param hdfs_driver: A string denoting the hdfs driver to use (if using a dataset on hdfs). Current choices are
        libhdfs (java through JNI) or libhdfs3 (C++)
        """
        self._dataset_url_parsed = urlparse(dataset_url)
        self._schema = schema
        self._ngram = ngram
        self._local_cache = local_cache
        self._worker_predicate = worker_predicate

        resolver = FilesystemResolver(self._dataset_url_parsed, hdfs_driver=hdfs_driver)
        self._dataset = pq.ParquetDataset(
            resolver.get_dataset_path(),
            filesystem=resolver.filesystem(),
            validate_schema=False)

    def load(self, rowgroup_spec):
        """Loads data form a single rowgroup from the dataset.

        Reads a single rowgroup from a dataset. Returns a list of dictionary with still encoded data.
        If worker_predicate was passed to the constructor, the predicate is first applied to the columns specified
        by the predicate. The rest of the columns are loaded only if at least one row matches the predicate.

        A rowgroup will be loaded from local cache, if cache contains an instance of the rowgroup.

        If ngram not None was passed to the constructor, the function returns a dictionary structured according to
        NGram definition.

        :param rowgroup_spec: A dictionary containing the following fields: 'row_group': ParquetDatasetPiece object
          describing a rowgroup to be loaded; 'shuffle_row_drop_partition' a tuple with
          (this_partition, num_of_partitions)
        :return: A dictionary indexed by field names, or a dictionary defined by NGram spec.
        """
        piece = rowgroup_spec['row_group']
        shuffle_row_drop_partition = rowgroup_spec['shuffle_row_drop_partition']

        # Create pyarrow file system
        with self._dataset.fs.open(piece.path) as piece_file_handle:
            parquet_file = ParquetFile(piece_file_handle)

            if not isinstance(self._local_cache, NullCache):
                if self._worker_predicate:
                    raise RuntimeError('Local cache is not supported together with predicates, '
                                       'unless the dataset is partitioned by the column the predicate operates on.')
                if shuffle_row_drop_partition[1] != 1:
                    raise RuntimeError('Local cache is not supported together with shuffle_row_drop_partitions > 1')

            if self._worker_predicate:
                all_cols = self._load_rows_with_predicate(parquet_file, piece, self._worker_predicate,
                                                          shuffle_row_drop_partition)
            else:
                # Using hash of the dataset url with the relative path in order to:
                #  1. Make sure if a common cache serves multiple processes (e.g. redis), we don't have conflicts
                #  2. Dataset url is hashed, to make sure we don't create too long keys, which maybe incompatible with
                #     some cache implementations
                #  3. Still leave relative path and the piece_index in plain text to make it easier to debug
                cache_key = '{}:{}:{}'.format(
                    hashlib.md5(urlunparse(self._dataset_url_parsed).encode('utf-8')).hexdigest(),
                    piece.path, piece.row_group)
                all_cols = self._local_cache.get(cache_key,
                                                 lambda: self._load_rows(parquet_file, piece,
                                                                         shuffle_row_drop_partition))

        if self._ngram:
            all_cols_as_ngrams = self._ngram.form_ngram(data=all_cols, schema=self._schema)
            return all_cols_as_ngrams
        else:
            return all_cols

    def _load_rows(self, parquet_file, piece, shuffle_row_drop_range):
        """Loads all rows from a piece"""

        # pyarrow would fail if we request a column names that the dataset is partitioned by, so we strip them from
        # the `columns` argument.
        partitions = self._dataset.partitions
        column_names = set(field.name for field in self._schema.fields.values()) - partitions.partition_names

        all_rows = self._read_with_shuffle_row_drop(piece, parquet_file, column_names, shuffle_row_drop_range)

        return all_rows

    def _load_rows_with_predicate(self, parquet_file, piece, worker_predicate, shuffle_row_drop_partition):
        """Loads all rows that match a predicate from a piece"""

        # 1. Read all columns needed by predicate and decode
        # 2. Apply the predicate. If nothing matches, exit early
        # 3. Read the remaining columns and decode
        # 4. Combine with columns already decoded for the predicate.

        # Split all column names into ones that are needed by predicateand the rest.
        predicate_column_names = set(worker_predicate.get_fields())

        if not predicate_column_names:
            raise ValueError('At least one field name must be returned by predicate\'s get_field() method')

        all_schema_names = set(field.name for field in self._schema.fields.values())

        invalid_column_names = predicate_column_names - all_schema_names
        if invalid_column_names:
            raise ValueError('At least some column names requested by the predicate ({}) '
                             'are not valid schema names: ({})'.format(', '.join(invalid_column_names),
                                                                       ', '.join(all_schema_names)))

        other_column_names = all_schema_names - predicate_column_names - \
            self._dataset.partitions.partition_names

        # Read columns needed for the predicate
        predicate_rows = self._read_with_shuffle_row_drop(piece, parquet_file, predicate_column_names,
                                                          shuffle_row_drop_partition)

        # Decode values
        decoded_predicate_rows = [_select_cols(row, predicate_column_names) for row in predicate_rows]

        # Use the predicate to filter
        match_predicate_mask = [worker_predicate.do_include(row) for row in decoded_predicate_rows]

        # Don't have anything left after filtering? Exit early.
        if not any(match_predicate_mask):
            return []

        # Remove rows that were filtered out by the predicate
        filtered_decoded_predicate_rows = [row for i, row in enumerate(decoded_predicate_rows) if
                                           match_predicate_mask[i]]

        if other_column_names:
            # Read remaining columns
            other_rows = self._read_with_shuffle_row_drop(piece, parquet_file, other_column_names,
                                                          shuffle_row_drop_partition)

            # Remove rows that were filtered out by the predicate
            filtered_other_rows = [row for i, row in enumerate(other_rows) if match_predicate_mask[i]]

            # Decode remaining columns
            decoded_other_rows = filtered_other_rows

            # Merge predicate needed columns with the remaining
            all_cols = [_merge_two_dicts(a, b) for a, b in zip(decoded_other_rows, filtered_decoded_predicate_rows)]
            return all_cols
        else:
            return filtered_decoded_predicate_rows

    def _read_with_shuffle_row_drop(self, piece, parquet_file, column_names, shuffle_row_drop_partition):
        data_frame = piece.read(
            open_file_func=lambda _: parquet_file,
            columns=column_names,
            partitions=self._dataset.partitions
        ).to_pandas(integer_object_nulls=True)

        num_rows = len(data_frame)
        num_partitions = shuffle_row_drop_partition[1]
        this_partition = shuffle_row_drop_partition[0]

        partition_indexes = np.floor(np.arange(num_rows) / (float(num_rows) / min(num_rows, num_partitions)))

        if self._ngram:
            # If we have an ngram we need to take elements from the next partition to build the ngram
            next_partition_indexes = np.where(partition_indexes >= this_partition + 1)
            if next_partition_indexes[0].size:
                next_partition_to_add = next_partition_indexes[0][0:self._ngram.length - 1]
                partition_indexes[next_partition_to_add] = this_partition

        selected_dataframe = data_frame.loc[partition_indexes == this_partition]
        return selected_dataframe.to_dict('records')



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import unittest
from decimal import Decimal

import numpy as np
import pyarrow as pa
import pytest
from pyspark import Row
from pyspark.sql.types import StringType, IntegerType, DecimalType, ShortType, LongType

from petastorm.codecs import ScalarCodec, NdarrayCodec
from petastorm.unischema import Unischema, UnischemaField, dict_to_spark_row, \
    insert_explicit_nulls, match_unischema_fields, _new_gt_255_compatible_namedtuple

try:
    from unittest import mock
except ImportError:
    from mock import mock


def _mock_parquet_dataset(partition_names, schema):
    """Creates a pyarrow.ParquetDataset mock capable of returning:

        parquet_dataset.pieces[0].get_metadata(parquet_dataset.fs.open).schema.to_arrow_schema() == schema
        parquet_dataset.partition_names.partition_names = partition_names

    """
    piece_mock = mock.Mock()
    piece_mock.get_metadata().schema.to_arrow_schema.return_value = schema

    dataset_mock = mock.Mock()
    type(dataset_mock).pieces = mock.PropertyMock(return_value=[piece_mock])
    type(dataset_mock.partitions).partition_names = mock.PropertyMock(return_value=partition_names)

    return dataset_mock


class UnischemaTest(unittest.TestCase):

    def test_fields(self):
        """Try using 'fields' getter"""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])

        self.assertEqual(len(TestSchema.fields), 2)
        self.assertEqual(TestSchema.fields['int_field'].name, 'int_field')
        self.assertEqual(TestSchema.fields['string_field'].name, 'string_field')

    def test_as_spark_schema(self):
        """Try using 'as_spark_schema' function"""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])

        spark_schema = TestSchema.as_spark_schema()
        self.assertEqual(spark_schema.fields[0].name, 'int_field')
        self.assertEqual(spark_schema.fields[1].name, 'string_field')

        self.assertEqual(TestSchema.fields['int_field'].name, 'int_field')
        self.assertEqual(TestSchema.fields['string_field'].name, 'string_field')

    def test_dict_to_spark_row_field_validation_scalar_types(self):
        """Test various validations done on data types when converting a dictionary to a spark row"""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])

        self.assertTrue(isinstance(dict_to_spark_row(TestSchema, {'string_field': 'abc'}), Row))

        # Not a nullable field
        with self.assertRaises(ValueError):
            isinstance(dict_to_spark_row(TestSchema, {'string_field': None}), Row)

        # Wrong field type
        with self.assertRaises(ValueError):
            isinstance(dict_to_spark_row(TestSchema, {'string_field': []}), Row)

    def test_dict_to_spark_row_field_validation_scalar_nullable(self):
        """Test various validations done on data types when converting a dictionary to a spark row"""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), True),
            UnischemaField('nullable_implicitly_set', np.string_, (), ScalarCodec(StringType()), True),
        ])

        self.assertTrue(isinstance(dict_to_spark_row(TestSchema, {'string_field': None}), Row))

    def test_dict_to_spark_row_field_validation_ndarrays(self):
        """Test various validations done on data types when converting a dictionary to a spark row"""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('tensor3d', np.float32, (10, 20, 30), NdarrayCodec(), False),
        ])

        self.assertTrue(isinstance(dict_to_spark_row(TestSchema,
                                                     {'tensor3d': np.zeros((10, 20, 30), dtype=np.float32)}), Row))

        # Null value into not nullable field
        with self.assertRaises(ValueError):
            isinstance(dict_to_spark_row(TestSchema, {'string_field': None}), Row)

        # Wrong dimensions
        with self.assertRaises(ValueError):
            isinstance(dict_to_spark_row(TestSchema, {'string_field': np.zeros((1, 2, 3), dtype=np.float32)}), Row)

    def test_make_named_tuple(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('string_scalar', np.string_, (), ScalarCodec(StringType()), True),
            UnischemaField('int32_scalar', np.int32, (), ScalarCodec(ShortType()), False),
            UnischemaField('uint8_scalar', np.uint8, (), ScalarCodec(ShortType()), False),
            UnischemaField('int32_matrix', np.float32, (10, 20, 3), NdarrayCodec(), True),
            UnischemaField('decimal_scalar', Decimal, (10, 20, 3), ScalarCodec(DecimalType(10, 9)), False),
        ])

        TestSchema.make_namedtuple(string_scalar='abc', int32_scalar=10, uint8_scalar=20,
                                   int32_matrix=np.int32((10, 20, 3)), decimal_scalar=Decimal(123) / Decimal(10))

        TestSchema.make_namedtuple(string_scalar=None, int32_scalar=10, uint8_scalar=20,
                                   int32_matrix=None, decimal_scalar=Decimal(123) / Decimal(10))

    def test_insert_explicit_nulls(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('nullable', np.int32, (), ScalarCodec(StringType()), True),
            UnischemaField('not_nullable', np.int32, (), ScalarCodec(ShortType()), False),
        ])

        # Insert_explicit_nulls to leave the dictionary as is.
        row_dict = {'nullable': 0, 'not_nullable': 1}
        insert_explicit_nulls(TestSchema, row_dict)
        self.assertEqual(len(row_dict), 2)
        self.assertEqual(row_dict['nullable'], 0)
        self.assertEqual(row_dict['not_nullable'], 1)

        # Insert_explicit_nulls to leave the dictionary as is.
        row_dict = {'nullable': None, 'not_nullable': 1}
        insert_explicit_nulls(TestSchema, row_dict)
        self.assertEqual(len(row_dict), 2)
        self.assertEqual(row_dict['nullable'], None)
        self.assertEqual(row_dict['not_nullable'], 1)

        # We are missing a nullable field here. insert_explicit_nulls should add a None entry.
        row_dict = {'not_nullable': 1}
        insert_explicit_nulls(TestSchema, row_dict)
        self.assertEqual(len(row_dict), 2)
        self.assertEqual(row_dict['nullable'], None)
        self.assertEqual(row_dict['not_nullable'], 1)

        # We are missing a not_nullable field here. Should raise an ValueError.
        row_dict = {'nullable': 0}
        with self.assertRaises(ValueError):
            insert_explicit_nulls(TestSchema, row_dict)

    def test_create_schema_view_fails_validate(self):
        """ Exercises code paths unischema.create_schema_view ValueError, and unischema.__str__."""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        with self.assertRaises(ValueError) as ex:
            TestSchema.create_schema_view([UnischemaField('id', np.int64, (), ScalarCodec(LongType()), False)])
        self.assertTrue('does not belong to the schema' in str(ex.exception))

    def test_create_schema_view_using_invalid_type(self):
        """ Exercises code paths unischema.create_schema_view ValueError, and unischema.__str__."""
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        with self.assertRaises(ValueError) as ex:
            TestSchema.create_schema_view([42])
        self.assertTrue('must be either a string' in str(ex.exception))

    def test_create_schema_view_using_unischema_fields(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        view = TestSchema.create_schema_view([TestSchema.int_field])
        self.assertEqual(set(view.fields.keys()), {'int_field'})

    def test_create_schema_view_using_regex(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        view = TestSchema.create_schema_view(['int.*$'])
        self.assertEqual(set(view.fields.keys()), {'int_field'})

        view = TestSchema.create_schema_view([u'int.*$'])
        self.assertEqual(set(view.fields.keys()), {'int_field'})

    def test_create_schema_view_using_regex_and_unischema_fields(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
            UnischemaField('other_string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        view = TestSchema.create_schema_view(['int.*$', TestSchema.string_field])
        self.assertEqual(set(view.fields.keys()), {'int_field', 'string_field'})

    def test_create_schema_view_no_field_matches_regex(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int_field', np.int8, (), ScalarCodec(IntegerType()), False),
            UnischemaField('string_field', np.string_, (), ScalarCodec(StringType()), False),
        ])
        view = TestSchema.create_schema_view(['bogus'])
        self.assertEqual(len(view.fields), 0)

    def test_name_property(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('nullable', np.int32, (), ScalarCodec(StringType()), True),
        ])

        self.assertEqual('TestSchema', TestSchema._name)

    def test_field_name_conflict_with_unischema_attribute(self):
        # fields is an existing attribute of Unischema
        with pytest.warns(UserWarning, match='Can not create dynamic property'):
            Unischema('TestSchema', [UnischemaField('fields', np.int32, (), ScalarCodec(StringType()), True)])

    def test_filter_schema_fields_from_url(self):
        TestSchema = Unischema('TestSchema', [
            UnischemaField('int32', np.int32, (), None, False),
            UnischemaField('uint8', np.uint8, (), None, False),
            UnischemaField('uint16', np.uint16, (), None, False),
        ])

        assert match_unischema_fields(TestSchema, ['.*nt.*6']) == [TestSchema.uint16]
        assert match_unischema_fields(TestSchema, ['nomatch']) == []
        assert match_unischema_fields(TestSchema, ['.*']) == list(TestSchema.fields.values())
        assert match_unischema_fields(TestSchema, ['int32', 'uint8']) == [TestSchema.int32, TestSchema.uint8]

    def test_arrow_schema_convertion(self):

        arrow_schema = pa.schema([
            pa.field('string', pa.string()),
            pa.field('int8', pa.int8()),
            pa.field('int16', pa.int16()),
            pa.field('int32', pa.int32()),
            pa.field('int64', pa.int64()),
            pa.field('float', pa.float32()),
            pa.field('double', pa.float64()),
            pa.field('bool', pa.bool_(), False),
            pa.field('fixed_size_binary', pa.binary(10)),
            pa.field('variable_size_binary', pa.binary()),
            pa.field('decimal', pa.decimal128(3, 4)),
            pa.field('timestamp_s', pa.timestamp('s')),
            pa.field('timestamp_ns', pa.timestamp('ns')),
            pa.field('date_32', pa.date32()),
            pa.field('date_64', pa.date64()),
            pa.field('timestamp_ns', pa.timestamp('ns')),
        ])

        mock_dataset = _mock_parquet_dataset([], arrow_schema)

        unischema = Unischema.from_arrow_schema(mock_dataset)
        for name in arrow_schema.names:
            assert getattr(unischema, name).name == name
            assert isinstance(getattr(unischema, name).codec, ScalarCodec)
            if name == 'bool':
                assert not getattr(unischema, name).nullable
            else:
                assert getattr(unischema, name).nullable

    def test_arrow_schema_convertion_with_partitions(self):

        arrow_schema = pa.schema([
            pa.field('int8', pa.int8()),
        ])

        mock_dataset = _mock_parquet_dataset(['part_name'], arrow_schema)

        unischema = Unischema.from_arrow_schema(mock_dataset)
        assert unischema.part_name.codec.spark_dtype().typeName() == 'string'

    def test_arrow_schema_convertion_fail(self):
        arrow_schema = pa.schema([
            pa.field('list_of_int', pa.float16()),
        ])

        mock_dataset = _mock_parquet_dataset([], arrow_schema)

        with self.assertRaises(ValueError) as ex:
            Unischema.from_arrow_schema(mock_dataset)

        assert 'Cannot auto-create unischema due to unsupported column type' in str(ex.exception)

    def test_arrow_schema_arrow_1644_list_of_struct(self):
        arrow_schema = pa.schema([
            pa.field('id', pa.string()),
            pa.field('list_of_struct', pa.list_(pa.struct([('a', pa.string()), ('b', pa.int32())])))
        ])

        mock_dataset = _mock_parquet_dataset([], arrow_schema)

        unischema = Unischema.from_arrow_schema(mock_dataset)
        assert getattr(unischema, 'id').name == 'id'
        assert not hasattr(unischema, 'list_of_struct')

    def test_arrow_schema_arrow_1644_list_of_list(self):
        arrow_schema = pa.schema([
            pa.field('id', pa.string()),
            pa.field('list_of_list',
                     pa.list_(pa.list_(pa.struct([('a', pa.string()), ('b', pa.int32())]))))
        ])

        mock_dataset = _mock_parquet_dataset([], arrow_schema)

        unischema = Unischema.from_arrow_schema(mock_dataset)
        assert getattr(unischema, 'id').name == 'id'
        assert not hasattr(unischema, 'list_of_list')

    def test_arrow_schema_convertion_ignore(self):
        arrow_schema = pa.schema([
            pa.field('list_of_int', pa.float16()),
            pa.field('struct', pa.struct([('a', pa.string()), ('b', pa.int32())])),
        ])

        mock_dataset = _mock_parquet_dataset([], arrow_schema)

        unischema = Unischema.from_arrow_schema(mock_dataset, omit_unsupported_fields=True)
        assert not hasattr(unischema, 'list_of_int')


class UnischemaFieldTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._TestField1a = UnischemaField('random', np.string_, (), ScalarCodec(StringType()), False)
        cls._TestField1b = UnischemaField('random', np.string_, (), ScalarCodec(StringType()), False)
        cls._TestField1c = UnischemaField('Random', np.string_, (), ScalarCodec(StringType()), False)
        cls._TestField2a = UnischemaField('id', np.int32, (), ScalarCodec(ShortType()), False)
        cls._TestField2b = UnischemaField('id', np.int32, (), ScalarCodec(ShortType()), False)
        cls._TestField2c = UnischemaField('ID', np.int32, (), ScalarCodec(ShortType()), False)

    def test_equality(self):
        # Use assertTrue instead of assertEqual/assertNotEqual so we don't depend on which operator (__eq__ or __ne__)
        # actual implementation of assert uses
        self.assertTrue(self._TestField1a == self._TestField1b)
        self.assertTrue(self._TestField2a == self._TestField2b)
        self.assertTrue(self._TestField1a != self._TestField1c)
        self.assertTrue(self._TestField2a != self._TestField2c)

    def test_hash(self):
        self.assertEqual(hash(self._TestField1a), hash(self._TestField1b))
        self.assertEqual(hash(self._TestField2a), hash(self._TestField2b))
        self.assertNotEqual(hash(self._TestField1a), hash(self._TestField1c))
        self.assertNotEqual(hash(self._TestField2a), hash(self._TestField2c))


def test_new_gt_255_compatible_namedtuple():
    fields_count = 1000
    field_names = ['f{}'.format(i) for i in range(fields_count)]
    values = list(range(1000))
    huge_tuple = _new_gt_255_compatible_namedtuple('HUGE_TUPLE', field_names)
    huge_tuple_instance = huge_tuple(**dict(zip(field_names, values)))
    assert len(huge_tuple_instance) == fields_count
    assert huge_tuple_instance.f764 == 764


if __name__ == '__main__':
    # Delegate to the test framework.
    unittest.main()



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from decimal import Decimal

import numpy as np
import pytest
from pyspark.sql.types import StringType, ByteType, ShortType, IntegerType, LongType, DecimalType

from petastorm.codecs import ScalarCodec
from petastorm.unischema import UnischemaField


def test_byte_string():
    codec = ScalarCodec(StringType())
    field = UnischemaField(name='field_string', numpy_dtype=np.string_, shape=(), codec=codec, nullable=False)

    assert codec.decode(field, codec.encode(field, 'abc')) == b'abc'
    assert codec.decode(field, codec.encode(field, '')) == b''


def test_unicode():
    codec = ScalarCodec(StringType())
    field = UnischemaField(name='field_string', numpy_dtype=np.unicode_, shape=(), codec=codec, nullable=False)

    assert codec.decode(field, codec.encode(field, 'abc')) == 'abc'
    assert codec.decode(field, codec.encode(field, '')) == ''


@pytest.mark.parametrize('spark_numpy_types', [
    (ByteType, np.uint8),
    (ByteType, np.int8),
    (ShortType, np.int16),
    (IntegerType, np.int32),
    (LongType, np.int64),
])
def test_numeric_types(spark_numpy_types):
    spark_type, numpy_type = spark_numpy_types

    codec = ScalarCodec(spark_type())
    field = UnischemaField(name='field_int', numpy_dtype=numpy_type, shape=(), codec=codec, nullable=False)

    min_val, max_val = np.iinfo(numpy_type).min, np.iinfo(numpy_type).max

    assert codec.decode(field, codec.encode(field, numpy_type(min_val))) == min_val
    assert codec.decode(field, codec.encode(field, numpy_type(max_val))) == max_val


def test_scalar_codec_decimal():
    codec = ScalarCodec(DecimalType(4, 3))
    field = UnischemaField(name='field_decimal', numpy_dtype=Decimal, shape=(), codec=codec, nullable=False)

    value = Decimal('123.4567')
    assert codec.decode(field, codec.encode(field, value)) == value


def test_bad_encoded_data_shape():
    codec = ScalarCodec(IntegerType())
    field = UnischemaField(name='field_int', numpy_dtype=np.int32, shape=(), codec=codec, nullable=False)
    with pytest.raises(TypeError):
        codec.decode(field, codec.encode(field, np.asarray([10, 10])))


def test_bad_unischema_field_shape():
    codec = ScalarCodec(IntegerType())
    field = UnischemaField(name='field_int', numpy_dtype=np.int32, shape=(1,), codec=codec, nullable=False)
    with pytest.raises(ValueError, match='must be an empty tuple'):
        codec.encode(field, np.int32(1))



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import logging
import threading
from collections import Counter
from concurrent.futures import ThreadPoolExecutor

import six
from pyarrow import parquet as pq
from six.moves.queue import Queue
from six.moves.urllib.parse import urlparse

from petastorm.cache import NullCache
from petastorm.etl import dataset_metadata, rowgroup_indexing
from petastorm.etl.dataset_metadata import infer_or_load_unischema
from petastorm.fs_utils import FilesystemResolver
from petastorm.ngram import NGram
from petastorm.predicates import PredicateBase
from petastorm.reader_impl.epochs import epoch_generator
from petastorm.reader_impl.row_group_decoder import RowDecoder
from petastorm.reader_impl.row_group_loader import RowGroupLoader
from petastorm.reader_impl.shuffling_buffer import NoopShufflingBuffer
from petastorm.reader_impl.worker_loop import worker_loop, EOFSentinel, WorkerLoopError
from petastorm.selectors import RowGroupSelectorBase

logger = logging.getLogger(__name__)

_OUTPUT_QUEUE_SIZE = 30


class ReaderV2(object):
    """Reads a unischema based dataset from a parquet file."""

    def __init__(self, dataset_url, schema_fields=None, predicate=None, rowgroup_selector=None,
                 num_epochs=1, sequence=None, cur_shard=None, shard_count=None,
                 read_timeout_s=None, cache=None, loader_pool=None, decoder_pool=None, shuffling_queue=None,
                 shuffle_row_groups=True, shuffle_row_drop_partitions=1, pyarrow_filesystem=None,
                 hdfs_driver='libhdfs3'):
        """Initializes a reader object.

        :param dataset_url: an filepath or a url to a parquet directory,
                       e.g. 'hdfs://some_hdfs_cluster/user/yevgeni/parquet8', or '/tmp/mydataset'
                       or ``'s3://bucket/mydataset'``.
        :param schema_fields:
            Either list of unischema fields to subset, or None to read all fields.
            OR an NGram object, then it will return an NGram of the specified properties.
        :param predicate: instance of predicate object to filter rows to be returned by reader.
        :param rowgroup_selector: instance of row group selector object to select row groups to be read
        :param reader_pool: parallelization pool. ThreadPool(10) (10 threads) is used by default.
                       This pool is a custom implementation used to parallelize reading data from the dataset.
                       Any object from workers_pool package can be used (e.g. ProcessPool)
        :param num_epochs: An epoch is a single pass over all samples in the dataset. Setting num_epochs to 'None' will
                       result in an infinite number of epochs.
        :param sequence: This is deprecated. To use sequence/ngram, please supply the argument in schema_fields instead.
        :param cur_shard: An int denoting the current shard number. Each node reading a shard should
                       pass in a unique shard number in the range [0, shard_count).
                       shard count must be supplied as well.
        :param shard_count An int denoting the number of shards to break this dataset into.
        :param read_timeout_s: A numeric with the amount of time in seconds you would like to give a read before it
                       times out and raises an EmptyResultError. Pass in None for an infinite timeout
        :param cache: An object conforming to `cache.CacheBase` interface. Before loading row groups from a parquet file
                       the Reader will attempt to load these values from cache. Caching is useful when communication
                       to the main data store is either slow or expensive and the local machine has large enough storage
                       to store entire dataset (or a partition of a dataset if num_training_partitions is used).
        :param decoder_pool: An instance of a concurrent.futures pool executor used for decoding. If None,
          a default ThreadPoolExecutor(5) will be used.
        :param loader_pool: An instance of a concurrent.futures pool executor used for decoding. If None,
          a default ThreadPoolExecutor(5) will be used.
        :param hdfs_driver: A string denoting the hdfs driver to use (if using a dataset on hdfs). Current choices are
        libhdfs (java through JNI) or libhdfs3 (C++)

        By default, `NullCache` implementation
        """

        # 1. Resolve dataset path (hdfs://, file://) and open the parquet storage (dataset)
        # 2. Get a list of all groups
        # 3. Filter rowgroups
        #    a. predicates
        #    b. row-group selector (our indexing mechanism)
        #    c. partition: used to get a subset of data for distributed training
        # 4. Launch a new thread running `worker_loop` function.

        if dataset_url is None or not isinstance(dataset_url, six.string_types):
            raise ValueError("""dataset_url must be a string""")

        if not (isinstance(schema_fields, collections.Iterable) or isinstance(schema_fields, NGram)
                or schema_fields is None):
            raise ValueError("""Fields must be either None, an iterable collection of Unischema fields or an NGram
            object.""")

        if sequence is not None:
            raise ValueError("""'sequence' argument of Reader object is deprecated. Please pass an NGram instance to
            'schema_fields' argument instead.""")

        # Can not rely on a check in epochs.py since it runs on a separate thread. Inform user earlier about invalid
        # argument value.
        if num_epochs is not None and (not isinstance(num_epochs, int) or num_epochs < 1):
            raise ValueError('iterations must be positive integer or None')

        self.ngram = schema_fields if isinstance(schema_fields, NGram) else None

        if self.ngram and not self.ngram.timestamp_overlap and shuffle_row_drop_partitions > 1:
            raise NotImplementedError('Using timestamp_overlap=False is not implemented with'
                                      ' shuffle_options.shuffle_row_drop_partitions > 1')

        cache = cache or NullCache()
        dataset_url = dataset_url[:-1] if dataset_url[-1] == '/' else dataset_url

        # 1. Resolve dataset path (hdfs://, file://) and open the parquet storage (dataset)
        logger.debug('dataset_url: %s', dataset_url)

        if pyarrow_filesystem is not None:
            filesystem = pyarrow_filesystem
            dataset_path = urlparse(dataset_url).path
        else:
            resolver = FilesystemResolver(dataset_url)
            filesystem = resolver.filesystem()
            dataset_path = resolver.get_dataset_path()

        self._dataset = pq.ParquetDataset(dataset_path, filesystem=filesystem, validate_schema=False)

        shuffle_row_drop_partitions = self._normalize_shuffle_options(shuffle_row_drop_partitions, self._dataset)

        stored_schema = infer_or_load_unischema(self._dataset)

        # Make a schema view (a view is a Unischema containing only a subset of fields
        # Will raise an exception if invalid schema fields are in schema_fields
        fields = schema_fields if isinstance(schema_fields, collections.Iterable) else None
        self.schema = stored_schema.create_schema_view(fields) if fields else stored_schema

        # 2. Get a list of all groups
        row_groups = dataset_metadata.load_row_groups(self._dataset)

        # 3. Filter rowgroups
        filtered_row_groups, worker_predicate = self._filter_row_groups(self._dataset, row_groups, predicate,
                                                                        rowgroup_selector, cur_shard,
                                                                        shard_count)

        epoch_items = self._apply_row_drop_partition(filtered_row_groups, shuffle_row_drop_partitions)

        # 4. Launch a new thread running `worker_loop` function.
        def epochs_iterator():
            return epoch_generator(epoch_items, num_epochs, shuffle_row_groups)

        self._results_queue = Queue(_OUTPUT_QUEUE_SIZE)

        loader = RowGroupLoader(dataset_url, self.schema, self.ngram, cache, worker_predicate, hdfs_driver=hdfs_driver)
        decoder = RowDecoder(self.schema, self.ngram)
        self._loader_pool = loader_pool or ThreadPoolExecutor(5)
        self._decoder_pool = decoder_pool or ThreadPoolExecutor(5)
        self._stop_flow_manager_event = threading.Event()
        self._diags = Counter()

        if not shuffling_queue:
            shuffling_queue = NoopShufflingBuffer()

        self._flow_manager_thread = threading.Thread(target=worker_loop,
                                                     args=(epochs_iterator, self._loader_pool, loader,
                                                           self._decoder_pool,
                                                           decoder,
                                                           shuffling_queue, self._results_queue,
                                                           self._stop_flow_manager_event, self._diags))
        self._flow_manager_thread.daemon = True
        self._flow_manager_thread.start()

        self._read_timeout_s = read_timeout_s
        self.batched_output = False

    def _apply_row_drop_partition(self, row_groups, shuffle_row_drop_partitions):
        items_to_ventilate = []
        for row_group in row_groups:
            for shuffle_row_drop_partition in range(shuffle_row_drop_partitions):
                items_to_ventilate.append(
                    {'row_group': row_group,
                     'shuffle_row_drop_partition': (shuffle_row_drop_partition,
                                                    shuffle_row_drop_partitions)})

        return items_to_ventilate

    def _filter_row_groups(self, dataset, row_groups, predicate, rowgroup_selector, training_partition,
                           num_training_partitions):
        """Calculates which rowgroups will be read during.

        The following filters are applied: predicates;  row-group selector (our indexing mechanism); training partition

        :param dataset: ParquetDataset instance
        :param row_groups: a list of row groups (a list of ParquetDatasetPiece objects)
        :param predicate: instance of predicate object to filter rows to be returned by reader.
        :param rowgroup_selector: instance of row group selector object to select row groups to be read
        :param training_partition: An int denoting the partition number used for multi node training. Each node should
                       pass in a unique partition number in the range [0, num_training_partitions).
                       num_training_partitions must be supplied as well.
        :param num_training_partitions An int denoting the number of training partitions (how many nodes are performing
                       the multi node training)
        :return: (filtered_row_group_indexes, worker_predicate): filtered_row_group_indexes an integer index into
        row_groups array. worker_predicate contains only predicates that could not be resolved on the partitioned fields
        and need to be evaluated by workers.
        """

        filtered_row_group_indexes, worker_predicate = \
            self._apply_predicate_to_row_groups(dataset, row_groups, predicate)

        if rowgroup_selector:
            filtered_row_group_indexes = self._apply_row_group_selector(dataset, rowgroup_selector,
                                                                        filtered_row_group_indexes)

        if training_partition is not None or num_training_partitions is not None:
            filtered_row_group_indexes = self._partition_row_groups(dataset, row_groups, num_training_partitions,
                                                                    training_partition,
                                                                    filtered_row_group_indexes)
        return [row_groups[i] for i in filtered_row_group_indexes], worker_predicate

    def _partition_row_groups(self, dataset, row_groups, num_training_partitions, training_partition,
                              filtered_row_group_indexes):
        """Filters the list of row group indexes based on the requested training partitions. Returns
        a modified list of rowgroup indexes."""

        if not num_training_partitions \
                or not isinstance(training_partition, int) \
                or not isinstance(num_training_partitions, int):
            raise ValueError('partition and num_partitions must be ints and both specified to use partitioning')

        # We hash on the relative path of each parquet file to guarantee consistency between different reader
        # constructions even after moving the dataset
        filtered_row_group_indexes = [index for index in filtered_row_group_indexes
                                      if index % num_training_partitions == training_partition]
        return filtered_row_group_indexes

    def _apply_row_group_selector(self, dataset, rowgroup_selector, filtered_row_group_indexes):
        """Filters the list of row group indexes using rowgroup selector object. Returns a modified list of rowgroup
        indexes."""

        if not isinstance(rowgroup_selector, RowGroupSelectorBase):
            raise ValueError('rowgroup_selector parameter is expected to be derived from RowGroupSelectorBase')

        # Load indexes from metadata
        available_row_group_indexes = rowgroup_indexing.get_row_group_indexes(dataset)

        required_indexes = rowgroup_selector.get_index_names()
        if not set(required_indexes).issubset(set(available_row_group_indexes.keys())):
            raise ValueError('Some of required indexes {} are not available in {}'.format(
                required_indexes, list(available_row_group_indexes.keys())))

        selected_indexes = rowgroup_selector.select_row_groups(available_row_group_indexes)

        # include only selected_indexes but in filtered_row_group_indexes order
        filtered_row_group_indexes = [idx for idx in filtered_row_group_indexes if idx in selected_indexes]
        return filtered_row_group_indexes

    def _apply_predicate_to_row_groups(self, dataset, row_groups, predicate):
        """Filters the list of row group indexes using rowgroup selector object. Returns a modified list of rowgroup
        indexes and a list of worker_predicate: predicates that could not be applied at this level
        (parquet partitioning)."""

        if predicate:
            if not isinstance(predicate, PredicateBase):
                raise ValueError('predicate parameter is expected to be derived from PredicateBase')
            predicate_fields = predicate.get_fields()

            if set(predicate_fields) == dataset.partitions.partition_names:
                assert len(dataset.partitions.partition_names) == 1, \
                    'Datasets with only a single partition level supported at the moment'

                filtered_row_group_indexes = []
                for piece_index, piece in enumerate(row_groups):
                    partition_name, partition_index = piece.partition_keys[0]
                    partition_value = dataset.partitions[0].keys[partition_index]
                    if predicate.do_include({partition_name: partition_value}):
                        filtered_row_group_indexes.append(piece_index)

                worker_predicate = None
            else:
                filtered_row_group_indexes = list(range(len(row_groups)))
                worker_predicate = predicate

        else:
            filtered_row_group_indexes = list(range(len(row_groups)))
            worker_predicate = None
        return filtered_row_group_indexes, worker_predicate

    @staticmethod
    def _normalize_shuffle_options(shuffle_row_drop_partitions, dataset):
        """Checks that shuffle_options doesnt ask for more patitions than rows in a row group.
        This prevents sending partitions to workers which will result in not reading anything."""
        if shuffle_row_drop_partitions > 1 and dataset.metadata and dataset.metadata.num_row_groups:
            max_rows_in_row_group = 1
            for i in six.moves.xrange(dataset.metadata.num_row_groups):
                max_rows_in_row_group = max(max_rows_in_row_group, dataset.metadata.row_group(i).num_rows)

            return min(shuffle_row_drop_partitions, max_rows_in_row_group)
        return shuffle_row_drop_partitions

    def stop(self):
        """Stops all worker threads/processes"""
        self._stop_flow_manager_event.set()

    def join(self):
        self._flow_manager_thread.join()
        self._loader_pool.shutdown()
        self._decoder_pool.shutdown(wait=False)

    @property
    def diagnostics(self):
        return self._diags

    def __iter__(self):
        return self

    def __next__(self):
        result = self._results_queue.get(timeout=self._read_timeout_s)
        if isinstance(result, EOFSentinel):
            raise StopIteration
        elif isinstance(result, WorkerLoopError):
            logger.error('An unhandled exception was raised in worker_loop: %s', str(result))
            raise result.inner_error
        return result

    def next(self):
        return self.__next__()

    # Functions needed to treat reader as a context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        self.join()



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from petastorm import utils


class RowDecoder(object):
    def __init__(self, schema, ngram):
        """RowDecoder decodes a row with encoded data loaded from a dataset and applies a decoder on each field.

        :param schema: A Unischema instance
        :param ngram: An NGram object instance. `ngram=None` indicates row is not an ngram
        """
        self._schema = schema
        self._ngram = ngram

    def decode(self, row):
        """Decodes each value in the dictionary as defined by Unischema (passed to the constructor). Returns a
        namedtuple

        Example (ngram is None):

        Input:
          row = {
            'field1': bytearray(),
            ...
          }

        Output:
          (
            field1=numpy.array(...),
            ...
          )


        Example (ngram):

        row = {
          -1 : { 'field1': bytearray(), ...},
          0 : { 'field2': bytearray(), ...},
          1 : { 'field1': bytearray(), ...},
        }

        Output:

        row = {
          -1 : ( field1=numpy.array(...), ...),
          0  : ( field2=numpy.array(...), ...),
          1  : ( field1=numpy.array(...), ...),
        }


        :param row: A dictionary of fields
        :return: A named tuple (or a dictionary of namedtuples in case of a non None ngram)
        """

        if self._ngram:
            for key in row.keys():
                current_schema = self._ngram.get_schema_at_timestep(self._schema, key)
                row[key] = current_schema.make_namedtuple(**utils.decode_row(row[key], current_schema))
            return row
        else:
            return self._schema.make_namedtuple(**utils.decode_row(row, self._schema))



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random


def epoch_generator(items, num_epochs, shuffle):
    """This generator is used to generate epochs.

    An epoch is a single pass on a set of data (without repetitions). If shuffle is False, items are yielded in
    the same order as passed into the generator. If shuffle is True, the items order is reshuffled for each epoch.
    If num_epochs is a None, infinite number of epochs will be generated. 'items' may contain any type of an
    object/primitive.

    Example:
      assert list(epoch_generator([1, 2, 3], 1, False)) == [1, 2, 3]
      assert list(epoch_generator([1, 2, 3], 2, False)) == [1, 2, 3, 1, 2, 3]

    :param items: List of objects in a single epoch.
    :param num_epochs: Number of epochs to generate
    :param shuffle: If True, the order of items in each epoch is randomized
    :return:
    """

    if num_epochs is not None and (not isinstance(num_epochs, int) or num_epochs < 1):
        raise ValueError('iterations must be positive integer or None')

    curr_item_index = 0

    epochs_left = num_epochs

    # Continue until 'epochs_left' is greater than 0. If 'None', continue forever
    while (epochs_left is None or epochs_left > 0) and items:

        if curr_item_index == 0 and shuffle:
            random.shuffle(items)

        yield items[curr_item_index]

        curr_item_index += 1

        if curr_item_index >= len(items):
            curr_item_index = 0

            # If iterations was set to None, that means we will iterate until stop is called
            if epochs_left is not None:
                epochs_left -= 1



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import as_completed

from petastorm.reader_impl.same_thread_executor import SameThreadExecutor


def test_single_future_submit():
    e = SameThreadExecutor()
    assert 10 == e.submit(lambda: 10).result()


def test_many_future_submit():
    e = SameThreadExecutor()
    futures = []
    for x in range(10):
        futures.append(e.submit(lambda y=x: y))

    completed = as_completed(futures)
    results = [f.result() for f in completed]
    assert list(range(10)) == sorted(results)


def test_max_workers():
    assert SameThreadExecutor()._max_workers == 1



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from petastorm import make_batch_reader

# pylint: disable=unnecessary-lambda
_D = [lambda url, **kwargs: make_batch_reader(url, reader_pool_type='dummy', **kwargs)]

# pylint: disable=unnecessary-lambda
_TP = [
    lambda url, **kwargs: make_batch_reader(url, reader_pool_type='thread', **kwargs),
    lambda url, **kwargs: make_batch_reader(url, reader_pool_type='process', **kwargs),
]


def _check_simple_reader(reader, expected_data):
    # Read a bunch of entries from the dataset and compare the data to reference
    expected_field_names = expected_data[0].keys()
    count = 0
    for row in reader:
        actual = row._asdict()

        # Compare value of each entry in the batch
        for i, id_value in enumerate(actual['id']):
            expected = next(d for d in expected_data if d['id'] == id_value)
            for field in expected_field_names:
                expected_value = expected[field]
                actual_value = actual[field][i, ...]
                np.testing.assert_equal(actual_value, expected_value)

        count += len(actual['id'])

    assert count == len(expected_data)


@pytest.mark.parametrize('reader_factory', _D + _TP)
def test_simple_read(scalar_dataset, reader_factory):
    """Just a bunch of read and compares of all values to the expected values using the different reader pools"""
    with reader_factory(scalar_dataset.url) as reader:
        _check_simple_reader(reader, scalar_dataset.data)


@pytest.mark.parametrize('reader_factory', _D)
def test_specify_columns_to_read(scalar_dataset, reader_factory):
    """Just a bunch of read and compares of all values to the expected values using the different reader pools"""
    with reader_factory(scalar_dataset.url, schema_fields=['id', 'float.*$']) as reader:
        sample = next(reader)
        assert set(sample._asdict().keys()) == {'id', 'float64'}
        assert sample.float64.size > 0


@pytest.mark.parametrize('reader_factory', _D)
def test_many_columns_non_petastorm_dataset(many_columns_non_petastorm_dataset, reader_factory):
    """Check if we can read a dataset with huge number of columns (1000 in this case)"""
    with reader_factory(many_columns_non_petastorm_dataset.url) as reader:
        sample = next(reader)
        assert set(sample._fields) == set(many_columns_non_petastorm_dataset.data[0].keys())

# TODO(yevgeni): missing tests: https://github.com/uber/petastorm/issues/257



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from petastorm.codecs import NdarrayCodec
from petastorm.unischema import UnischemaField, Unischema
from petastorm.utils import decode_row, DecodeFieldError

MatrixField = UnischemaField('matrix', np.float64, (10, 10), NdarrayCodec(), False)
MatrixSchema = Unischema('TestSchema', [MatrixField])


def test_nominal_case():
    """Nominal flow: can decode field successfully"""
    expected = np.random.rand(10, 10)
    row = {'matrix': NdarrayCodec().encode(MatrixField, expected)}

    actual = decode_row(row, MatrixSchema)['matrix']

    np.testing.assert_equal(actual, expected)


def test_can_not_decode():
    """Make sure field name is part of the error message"""
    row = {'matrix': 'bogus'}

    with pytest.raises(DecodeFieldError, match='matrix'):
        decode_row(row, MatrixSchema)



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import concurrent.futures
import logging
import sys
import traceback
from collections import Counter
from time import sleep
from traceback import format_exception

from six.moves import queue

logger = logging.getLogger(__name__)

_MAX_IN_LOADING_QUEUE_SIZE = 14
_MAX_IN_DECODING_QUEUE_SIZE = 14
_LOOP_IDLE_TIME_SEC = 0.001  # 1 ms


class EOFSentinel(object):
    """This is a 'poison' object being queued into the output queue to signal that no further data will be arriving.
    This would happen when the number of epochs is limited and we are done reading all the data."""


def _decode_row_dispatcher(decoder, row):
    return decoder.decode(row)


class WorkerLoopError(Exception):
    def __init__(self, inner_error, exc_info):
        self.inner_error = inner_error
        self.exc_info = exc_info

    def __str__(self):
        return 'worker_loop has failed with {}. Call stack:\n{}'.format(self.inner_error,
                                                                        ''.join(format_exception(*self.exc_info)))


# Runs in a separate thread
def worker_loop(epochs_generator, loading_pool, loader, decoding_pool, decoder, shuffling_queue, output_queue,
                stop_event, stats):
    # Possibly infinite loop driven by row-group ventilation logic (infinite if epochs=inf)
    in_loading_futures = []
    in_decoding = []

    if not isinstance(stats, Counter):
        raise ValueError('stats argument is expected to be a collections.Counter')

    try:
        epochs_iterator = iter(epochs_generator())
        # rowgroup_spec is currently a ParquetDatasetPiece, but will likely be augmented with petastorm specific info
        # in the future
        try:
            rowgroup_spec = next(epochs_iterator)
        except StopIteration:
            rowgroup_spec = None

        # Continue iterating until we have some data in one of the queues connecting loader/shuffling-queue/decoder or
        # we still have some rowgroup_specs to process
        while rowgroup_spec is not None or in_loading_futures or in_decoding or shuffling_queue.size > 0:
            # An owner can signal a stop before we are actually finished
            if stop_event.is_set():
                break

            # 1. Read more rowgroup_specs from our epochs generator and schedule 'loader' to load these rowgroups
            # 2. Readout what was already loaded and enqueue loaded rows into a shuffling queue
            # 3. Read from the shuffling queue and submit decoding tasks to decoding executor
            # 4. Read completed decoding futures and put the results into the final output queue

            # 1. Read more rowgroup_specs from our epochs generator and schedule 'loader' to load these rowgroups
            # ---------------------------------------------------------------------------------------------------

            # Need to keep a cap (_MAX_IN_LOADING_QUEUE_SIZE) on the queue so we don't get unbounded growth with
            # infinite epochs.
            if rowgroup_spec is not None and len(in_loading_futures) < _MAX_IN_LOADING_QUEUE_SIZE:
                in_loading_futures.append(loading_pool.submit(loader.load, rowgroup_spec))
                stats['rowgroups_scheduled_for_loading'] += 1
                stats['in_loading_queue_len'] = len(in_loading_futures)

                try:
                    rowgroup_spec = next(epochs_iterator)
                except StopIteration:
                    rowgroup_spec = None

            # 2. Readout what was already loaded and enqueue loaded rows into a shuffling queue
            # ---------------------------------------------------------------------------------

            # Futures that have completed will be placed first in the returned list
            loaded_futures = concurrent.futures.as_completed(in_loading_futures, timeout=0)

            # If no more data is ever expected from the loading, we should let shuffling buffer
            # deplete itself. Otherwise, decoding will wait forever for data from shuffling that
            # will never come.
            if not in_loading_futures and not rowgroup_spec:
                shuffling_queue.finish()

            # Push all loaded_futures and partially decoded data into shuffling queue
            while shuffling_queue.can_add():
                try:
                    loaded_future = next(loaded_futures)
                    loaded_future_result = loaded_future.result()
                    stats['rows_loaded'] += len(loaded_future_result)
                    stats['shuffling_buffer_size'] = shuffling_queue.size
                    shuffling_queue.add_many(loaded_future_result)

                    # Done processing, remove from the queue. 'remove' on the list is ok since we cap the list
                    # to a small size with _MAX_IN_LOADING_QUEUE_SIZE
                    in_loading_futures.remove(loaded_future)
                except (concurrent.futures.TimeoutError, StopIteration):
                    break

            # 3. Read from the shuffling queue and submit decoding tasks to decoding executor
            # ---------------------------------------------------------------------------------
            while shuffling_queue.can_retrieve() and len(in_decoding) < _MAX_IN_DECODING_QUEUE_SIZE:
                shuffled = shuffling_queue.retrieve()
                stats['rows_read_from_shuffling_queue'] += 1
                in_decoding.append(decoding_pool.submit(_decode_row_dispatcher, decoder, shuffled))
                stats['in_decoding_queue_len'] = len(in_decoding)

            # 4. Read completed decoding futures and put the results into the final output queue
            # ---------------------------------------------------------------------------------

            # We want to keep the order in the output queue the same as in the decoding input. This determinism
            # is useful for testing purposes and if we actually want to read in order (need to use
            # SingleThreadExecutor for that though)
            decoded = concurrent.futures.as_completed(in_decoding, timeout=0)

            # We'll keep a list of finished futures (already_decoded) in a set, but would follow the order of futures
            # in the `in_decoding` list when writing to output_queue

            already_decoded = set()
            while True:
                try:
                    next_decoded = next(decoded)
                    stats['rows_read_from_decoder'] += 1
                    already_decoded.add(next_decoded)
                except (concurrent.futures.TimeoutError, StopIteration):
                    break

            # follow the order of futures in the `in_decoding` list
            updated_in_decoding = []
            for i, was_decoding in enumerate(in_decoding):
                if was_decoding in already_decoded:
                    stats['rows_written_to_output_queue'] += 1
                    while not stop_event.is_set():
                        try:
                            output_queue.put(was_decoding.result(), block=False)
                            break
                        except queue.Full:
                            sleep(_LOOP_IDLE_TIME_SEC)
                else:
                    updated_in_decoding.append(in_decoding[i])

            in_decoding = updated_in_decoding
            stats['output_queue_size'] = output_queue.qsize()
            sleep(_LOOP_IDLE_TIME_SEC)

        # If we were ordered to stop, better not write the sentinel out since the queue can be full
        # and we would block forever. There should not be an expectation from the caller to get a sentinel in this case
        if not stop_event.is_set():
            output_queue.put(EOFSentinel())

    except Exception as e:  # pylint: disable=broad-except
        logger.error(traceback.format_exc())
        output_queue.put(WorkerLoopError(e, sys.exc_info()))



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A mock executor. Guarantees stable order of future evaluations: a submitted future is
evaluated immediately as part of submit implementation."""

from concurrent.futures import Executor
from concurrent.futures._base import Future


class SameThreadExecutor(Executor):
    """A mock executor. Guarantees stable order of future evaluations: a submitted future is
    evaluated immediately as part of submit implementation."""

    def __init__(self):
        super(SameThreadExecutor, self).__init__()

        # For now we define _max_workers and polymorphically with ThreadPoolExecutor and ProcessPoolExecutor in
        # worker_loop. We should not rely on private _max_workers for that, but first we need to refactor
        # ReaderV2 so we would know during construction of the reader what is the number of workers. Currently we
        # just get an *PoolExecutor instance which don't have a public API for that.
        self._max_workers = 1

    def submit(self, fn, *args, **kwargs):
        future = Future()
        future.set_result(fn(*args, **kwargs))
        return future



import pytest

from petastorm.reader_impl.epochs import epoch_generator


def test_one_epoch():
    assert list(epoch_generator([1, 2, 3], 1, False)) == [1, 2, 3]


def test_three_epochs():
    assert list(epoch_generator([1, 2, 3], 3, False)) == [1, 2, 3, 1, 2, 3, 1, 2, 3]


def test_shuffle():
    items_in_epoch = 50
    shuffled_epoch = list(epoch_generator(list(range(items_in_epoch)), 1, True))
    assert len(shuffled_epoch) == items_in_epoch
    assert shuffled_epoch != range(items_in_epoch)


def test_shuffle_per_epoch():
    items_in_epoch = 50
    two_epochs = list(epoch_generator(list(range(items_in_epoch)), 2, True))
    assert len(two_epochs) == 100
    assert two_epochs[:50] != two_epochs[50:]


def test_empty_items_list():
    assert not list(epoch_generator([], 2, True))
    assert not list(epoch_generator([], None, True))


def test_item_type_can_be_anything():
    assert list(epoch_generator([None, 'a string', {}], 1, False)) == [None, 'a string', {}]


def test_unlimited_epochs():
    items = [1, 2, 3]
    inf_epochs = epoch_generator(items, None, True)
    for _ in range(1000):
        actual_item = next(inf_epochs)
        assert actual_item in items


def test_one_item_epoch():
    assert list(epoch_generator([1, ], 3, False)) == [1, 1, 1]


def test_invalid_number_of_epochs():
    with pytest.raises(ValueError):
        list(epoch_generator([0, 1], 0, False))

    with pytest.raises(ValueError):
        list(epoch_generator([0, 1], -1, False))

    with pytest.raises(ValueError):
        list(epoch_generator([0, 1], 'abc', False))



#  Copyright (c) 2017-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Must import pyarrow before torch. See: https://github.com/uber/petastorm/blob/master/docs/troubleshoot.rst
import re

import pyarrow  # noqa: F401 pylint: disable=W0611

import collections
import decimal

import numpy as np
from six import PY2
from torch.utils.data.dataloader import default_collate

if PY2:
    _string_classes = basestring  # noqa: F821
else:
    _string_classes = (str, bytes)


def _sanitize_pytorch_types(row_as_dict):
    """Promotes values types in a dictionary to the types supported by pytorch. Raises an error if type is clear error
    if the type can not be promoted.

    The parameter is modified in-place.

    int8, uint16 are promoted to int32; uint32 -> int64;
    numpy string_, unicode_, object arrays are not supported.

    :param dict[str,obj] row_as_dict: a dictionary of key-value pairs. The values types are promoted to
        pytorch compatible.
    :return: None
    """
    for name, value in row_as_dict.items():
        # PyTorch supported types are: double, float, float16, int64, int32, and uint8
        if isinstance(value, np.ndarray):
            if value.dtype == np.int8:
                row_as_dict[name] = value.astype(np.int16)
            elif value.dtype == np.uint16:
                row_as_dict[name] = value.astype(np.int32)
            elif value.dtype == np.uint32:
                row_as_dict[name] = value.astype(np.int64)
            elif value.dtype == np.bool_:
                row_as_dict[name] = value.astype(np.uint8)
            elif re.search('[SaUO]', value.dtype.str):
                raise TypeError('Pytorch does not support arrays of string classes. Found in field {}'.format(name))
        elif isinstance(value, np.bool_):
            row_as_dict[name] = np.uint8(value)
        elif value is None:
            raise TypeError('Pytorch does not support nullable fields. Found None in {}'.format(name))


def decimal_friendly_collate(batch):
    """A wrapper on top of ``default_collate`` function that allows decimal.Decimal types to be collated.

    We use ``decimal.Decimal`` types in petastorm dataset to represent timestamps. PyTorch's ``default_collate``
    implementation does not support collating ``decimal.Decimal`` types. ``decimal_friendly_collate`` collates
    ``decimal.Decimal`` separately and then combines with the rest of the fields collated by a standard
    ``default_collate``.

    :param batch: A list of dictionaries to collate
    :return: A dictionary of lists/pytorch.Tensor types
    """

    if isinstance(batch[0], decimal.Decimal):
        return batch
    elif isinstance(batch[0], collections.Mapping):
        return {key: decimal_friendly_collate([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], _string_classes):
        return batch
    elif isinstance(batch[0], collections.Sequence):
        transposed = zip(*batch)
        return [decimal_friendly_collate(samples) for samples in transposed]
    else:
        return default_collate(batch)


class DataLoader(object):
    """
    A data loader adaptor for ``torch.utils.data.DataLoader``.

    This class iterates and returns items from the Reader in batches.

    This loader can be used as an iterator and will terminate when the reader used in the construction of the class
    runs out of samples.
    """

    def __init__(self, reader, batch_size=1, collate_fn=decimal_friendly_collate):
        """
        Initializes a data loader object, with a default collate.

        Number of epochs is defined by the configuration of the reader argument.

        :param reader: petastorm Reader instance
        :param batch_size: the number of items to return per batch; factored into the len() of this reader
        :param collate_fn: an optional callable to merge a list of samples to form a mini-batch.
        """
        self.reader = reader
        self.batch_size = batch_size
        self.collate_fn = collate_fn

    def __iter__(self):
        """
        The Data Loader iterator stops the for-loop when reader runs out of samples.
        """
        batch = []
        for row in self.reader:
            # Default collate does not work nicely on namedtuples and treat them as lists
            # Using dict will result in the yielded structures being dicts as well
            row_as_dict = row._asdict()
            _sanitize_pytorch_types(row_as_dict)
            batch.append(row_as_dict)
            if len(batch) == self.batch_size:
                yield self.collate_fn(batch)
                batch = []
        if batch:
            yield self.collate_fn(batch)

    # Functions needed to treat data loader as a context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reader.stop()
        self.reader.join()
